from pathlib import Path
import typing
import urllib.request
import urllib.error
import socket


def download(odir: Path, source_url: str, irng: typing.Sequence[int]):
    """Download star index files.
    The default range was useful for my cameras.
    """
    assert len(irng) == 2, "specify start, stop indices"

    odir = Path(odir).expanduser()
    odir.mkdir(parents=True, exist_ok=True)

    ri = int(source_url.split("/")[-2][:2])

    for i in range(*irng):
        fn = f"index-{ri:2d}{i:02d}.fits"
        url = f"{source_url}{fn}"
        ofn = odir / fn
        if ofn.is_file():  # no clobber
            print("Exists:", ofn)
            continue
        print(f"{url} => {ofn}")

        url_retrieve(url, ofn)


def url_retrieve(url: str, outfile: Path, overwrite: bool = False):
    """
    Parameters
    ----------
    url: str
        URL to download from
    outfile: pathlib.Path
        output filepath (including name)
    overwrite: bool
        overwrite if file exists
    """
    outfile = Path(outfile).expanduser().resolve()
    if outfile.is_dir():
        raise ValueError("Please specify full filepath, including filename")
    # need .resolve() in case intermediate relative dir doesn't exist
    if overwrite or not outfile.is_file():
        outfile.parent.mkdir(parents=True, exist_ok=True)
        try:
            urllib.request.urlretrieve(url, str(outfile))
        except (socket.gaierror, urllib.error.URLError) as err:
            raise ConnectionError("could not download {} due to {}".format(url, err))



#!/usr/bin/env python
"""
Astrometry-azel
Copyright (C) 2013-2018 Michael Hirsch, Ph.D.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

--------------------------------------
Image stack -> average -> write FITS


Because ImageJ has been a little buggy about writing FITS files, in particular the header
that astrometry.net then crashes on, we wrote this quick script to ingest a variety
of files and average the specified frames then write a FITS.
The reason we average a selected stack of images is to reduce the noise for use in
astrometry.net

The error you might get from an ImageJ saved FITS when reading in:
PyFits, AstroPy, or ImageMagick is:
IOError: Header missing END card.
"""
from pathlib import Path
import numpy as np
from typing import Tuple, Optional
from datetime import datetime
from astropy.io import fits
import logging

try:
    import imageio
except ImportError:
    imageio = None
try:
    import h5py
except ImportError:
    h5py = None
try:
    from scipy.io import loadmat
except ImportError:
    loadmat = None


def meanstack(
    infn: Path, Navg: int, ut1: Optional[datetime] = None, method: str = "mean"
) -> Tuple[np.ndarray, Optional[datetime]]:

    infn = Path(infn).expanduser().resolve(strict=True)
    # %% parse indicies to load
    if isinstance(Navg, slice):
        key = Navg
    elif isinstance(Navg, int):
        key = slice(0, Navg)
    elif len(Navg) == 1:
        key = slice(0, Navg[0])
    elif len(Navg) == 2:
        key = slice(Navg[0], Navg[1])
    else:
        raise ValueError(f"not sure what you mean by Navg={Navg}")
    # %% load images
    """
    some methods handled individually to improve efficiency with huge files
    """
    if infn.suffix == ".h5":
        if h5py is None:
            raise ImportError("pip install h5py")
        img, ut1 = _h5mean(infn, ut1, key, method)
    elif infn.suffix in (".fits", ".new"):
        # mmap doesn't work with BZERO/BSCALE/BLANK
        with fits.open(infn, mode="readonly", memmap=False) as f:
            img = collapsestack(f[0].data, key, method)
    elif infn.suffix == ".mat":
        if loadmat is None:
            raise ImportError("pip install scipy")
        img = loadmat(infn)
        img = collapsestack(img["data"].T, key, method)  # matlab is fortran order
    else:  # .tif etc.
        if imageio is None:
            raise ImportError("pip install imageio")
        img = imageio.imread(infn, as_gray=True)
        if img.ndim in (3, 4) and img.shape[-1] == 3:  # assume RGB
            img = collapsestack(img, key, method)

    return img, ut1


def _h5mean(
    fn: Path, ut1: Optional[datetime], key: slice, method: str
) -> Tuple[np.ndarray, Optional[datetime]]:
    with h5py.File(fn, "r") as f:
        img = collapsestack(f["/rawimg"], key, method)
        # %% time
        if ut1 is None:
            try:
                ut1 = f["/ut1_unix"][key][0]
            except KeyError:
                pass
        # %% orientation
        try:
            img = np.rot90(img, k=f["/params"]["rotccw"])
        except KeyError:
            pass

    return img, ut1


def collapsestack(img: np.ndarray, key: slice, method: str) -> np.ndarray:
    if img.ndim not in (2, 3):
        raise ValueError("only 2D or 3D image stacks are handled")

    # %% 2-D
    if img.ndim == 2:
        return img
    # %% 3-D
    if method == "mean":
        func = np.mean
    elif method == "median":
        func = np.median
    else:
        raise TypeError(f"unknown method {method}")

    return func(img[key, ...], axis=0).astype(img.dtype)


def writefits(img: np.ndarray, outfn: Path):
    outfn = Path(outfn).expanduser()

    f = fits.PrimaryHDU(img)
    try:
        f.writeto(outfn, overwrite=False, checksum=True)
        # no close()
        print("writing", outfn)
    except OSError:
        logging.warning(f"did not overwrite existing {outfn}")


def readh5coord(fn: Path) -> Tuple[float, float]:
    if not fn.suffix == ".h5":
        return None

    with h5py.File(fn, "r") as f:
        try:
            latlon = (f["/sensorloc"]["glat"], f["/sensorloc"]["glon"])
        except KeyError:
            try:
                latlon = f["/lla"][:2]
            except KeyError:
                return None

    return latlon



from pathlib import Path
import subprocess
import shutil
import logging
import warnings
from dateutil.parser import parse
from datetime import datetime
from typing import Tuple
from numpy import meshgrid, column_stack
import xarray
from astropy.io import fits  # instead of obsolete pyfits
from astropy.wcs import wcs

try:
    import pymap3d
except ImportError:
    pymap3d = None


def fits2radec(
    fitsfn: Path, WCSfn: Path = None, solve: bool = False, args: str = None
) -> xarray.Dataset:

    fitsfn = Path(fitsfn).expanduser()

    if WCSfn is None:
        if fitsfn.suffix in (".fits", ".new"):
            # using .wcs will also work but gives a spurious warning
            WCSfn = fitsfn.with_suffix(".wcs")
        elif fitsfn.suffix == ".wcs":
            WCSfn = fitsfn
        else:
            raise ValueError(f"please convert {fitsfn} to GRAYSCALE .fits")

    if solve:
        doSolve(fitsfn, args)

    with fits.open(fitsfn, mode="readonly") as f:
        yPix, xPix = f[0].shape[-2:]

    x, y = meshgrid(range(xPix), range(yPix))  # pixel indices to find RA/dec of
    xy = column_stack((x.ravel(order="C"), y.ravel(order="C")))
    # %% use astropy.wcs to register pixels to RA/DEC
    """
    http://docs.astropy.org/en/stable/api/astropy.wcs.WCS.html#astropy.wcs.WCS
    naxis=[0,1] is to take x,y axes in case a color photo was input e.g. to astrometry.net cloud solver
    """
    try:
        with fits.open(WCSfn, mode="readonly") as f:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                radec = wcs.WCS(f[0].header).all_pix2world(xy, 0)
                # radec = wcs.WCS(hdul[0].header,naxis=[0,1]).all_pix2world(xy, 0)
    except OSError:
        raise OSError(
            f"It appears the WCS solution is not present, was the FITS image solved?  looking for: {WCSfn}"
        )

    ra = radec[:, 0].reshape((yPix, xPix), order="C")
    dec = radec[:, 1].reshape((yPix, xPix), order="C")
    # %% collect output
    radec = xarray.Dataset(
        {"ra": (("y", "x"), ra), "dec": (("y", "x"), dec)},
        {"x": range(xPix), "y": range(yPix)},
        attrs={"filename": str(fitsfn)},
    )

    return radec


def radec2azel(
    scale: xarray.Dataset, latlon: Tuple[float, float], time: datetime
) -> xarray.Dataset:

    if pymap3d is None:
        logging.error("azimuth, elevation computations require: pip install pymap3d")
        return None
    if latlon is None:
        return None
    if not isinstance(scale, xarray.Dataset):
        return None

    if time is None:
        with fits.open(scale.filename, mode="readonly") as f:
            try:
                t = f[0].header["FRAME"]  # TODO this only works from Solis?
            except KeyError:
                return None
        time = parse(t)
        logging.info("using FITS header for time")
    elif isinstance(time, datetime):
        pass
    elif isinstance(time, (float, int)):  # assume UT1_Unix
        time = datetime.utcfromtimestamp(time)
    else:  # user override of frame time
        time = parse(time)

    print("image time:", time)
    # %% knowing camera location, time, and sky coordinates observed, convert to az/el for each pixel
    # .values is to avoid silently freezing AstroPy
    az, el = pymap3d.radec2azel(scale["ra"].values, scale["dec"].values, *latlon, time)
    if (el < 0).any():
        Nbelow = (el < 0).nonzero()
        logging.error(
            f"{Nbelow} points were below the horizon."
            "Currently this program assumed observer ~ ground level."
            "Please file a bug report if you need observer off of Earth surface"
        )
    # %% collect output
    scale["az"] = (("y", "x"), az)
    scale["el"] = (("y", "x"), el)
    scale.attrs["lat"] = latlon[0]
    scale.attrs["lon"] = latlon[1]
    scale.attrs["time"] = time
    return scale


def doSolve(fitsfn: Path, args: str = None):
    """
    Astrometry.net from at least version 0.67 is OK with Python 3.
    """
    solve = shutil.which("solve-field")
    if not solve:
        raise FileNotFoundError("Astrometry.net solve-file exectuable not found")

    if isinstance(args, str):
        opts = args.split(" ")
    elif args is None:
        args = []
    # %% build command
    cmd = [solve, "--overwrite", str(fitsfn)]
    cmd += opts
    print("\n", " ".join(cmd), "\n")
    # %% execute
    ret = subprocess.check_output(cmd, universal_newlines=True)

    # solve-field returns 0 even if it didn't solve!
    print(ret)
    if "Did not solve" in ret:
        raise RuntimeError(f"could not solve {fitsfn}")

    print("\n\n*** done with astrometry.net ***\n")


def fits2azel(
    fitsfn: Path,
    *,
    wcsfn: Path = None,
    latlon: Tuple[float, float] = None,
    time: datetime = None,
    solve: bool = False,
    args: str = None,
) -> xarray.Dataset:

    fitsfn = Path(fitsfn).expanduser()

    radec = fits2radec(fitsfn, wcsfn, solve, args)
    # if az/el can be computed, scale is implicitly merged with radec.
    scale = radec2azel(radec, latlon, time)
    if scale is None:
        scale = radec

    return scale



from pathlib import Path
import logging
import xarray
from matplotlib.pyplot import figure
import numpy as np
from typing import Sequence


def plotazel(scale: xarray.Dataset, plottype: str = "contour", img: np.ndarray = None):

    if "az" not in scale:
        return

    fg = figure(figsize=(12, 5))
    ax = fg.subplots(1, 2, sharey=True)

    fg.suptitle(
        f"{Path(scale.filename).name}\n({scale.lat:.2f}, {scale.lon:.2f})  {scale.time}"
    )
    # %%
    axa = ax[0]

    if plottype == "image":
        hia = axa.imshow(scale["az"], origin="lower")
        hc = fg.colorbar(hia)
        hc.set_label("Azimuth [deg]")
    elif plottype == "contour":
        if img is not None:
            axa.imshow(img, origin="lower", cmap="gray")
        cs = axa.contour(scale["x"], scale["y"], scale["az"])
        axa.clabel(cs, inline=1, fmt="%0.1f")

    axa.set_xlabel("x-pixel")
    axa.set_ylabel("y-pixel")
    axa.set_title("azimuth")
    # %%
    axe = ax[1]
    if plottype == "image":
        hie = axe.imshow(scale["el"], origin="lower")
        hc = fg.colorbar(hie)
        hc.set_label("Elevation [deg]")
    elif plottype == "contour":
        if img is not None:
            axe.imshow(img, origin="lower", cmap="gray")
        cs = axe.contour(scale["x"], scale["y"], scale["el"])
        axe.clabel(cs, inline=True, fmt="%0.1f", fontsize="medium")

    axe.set_xlabel("x-pixel")
    axe.set_title("elevation")
    fg.set_tight_layout(True)

    return fg


def plotradec(scale: xarray.Dataset, plottype: str = "contour", img: np.ndarray = None):

    if "ra" not in scale:
        return

    fg = figure(figsize=(12, 5))
    axs = fg.subplots(1, 2, sharey=True)

    fg.suptitle(Path(scale.filename).name)

    ax = axs[0]

    if plottype == "image":
        hri = ax.imshow(scale["ra"], origin="lower")
        hc = fg.colorbar(hri)
        hc.set_label("RA [deg]")
    elif plottype == "contour":
        if img is not None:
            ax.imshow(img, origin="lower", cmap="gray")
        cs = ax.contour(scale["x"], scale["y"], scale["ra"])
        ax.clabel(cs, inline=1, fmt="%0.1f")

    ax.set_xlabel("x-pixel")
    ax.set_ylabel("y-pixel")
    ax.set_title("Right Ascension ")
    # %%
    ax = axs[1]
    if plottype == "image":
        hdi = ax.imshow(scale["dec"], origin="lower")
        hc = fg.colorbar(hdi)
        hc.set_label("Dec [deg]")
    elif plottype == "contour":
        if img is not None:
            ax.imshow(img, origin="lower", cmap="gray")
        cs = ax.contour(scale["x"], scale["y"], scale["dec"])
        ax.clabel(cs, inline=1, fmt="%0.1f")

    ax.set_xlabel("x-pixel")
    ax.set_title("Declination")
    fg.set_tight_layout(True)

    return fg


def plotimagestack(img: np.ndarray, fn: Path, makeplot: Sequence[str], clim=None):
    fn = Path(fn).expanduser()
    # %% plotting
    if img.ndim == 3 and img.shape[0] == 3:  # it seems to be an RGB image
        cmap = None
        imnorm = None
        img = img.transpose([1, 2, 0])  # imshow() needs colors to be last axis
    else:
        cmap = "gray"
        imnorm = None
        # imnorm = LogNorm()

    fg = figure()
    ax = fg.gca()
    if clim is None:
        hi = ax.imshow(
            img, origin="lower", interpolation="none", cmap=cmap, norm=imnorm
        )
    else:
        hi = ax.imshow(
            img,
            origin="lower",
            interpolation="none",
            cmap=cmap,
            vmin=clim[0],
            vmax=clim[1],
            norm=imnorm,
        )
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_title(str(fn))
    if cmap is not None:
        try:
            hc = fg.colorbar(hi)
            hc.set_label(f"Data numbers {img.dtype}")
        except Exception as e:
            logging.warning(f"trouble making picture colorbar  {e}")
    # %%
    if "png" in makeplot:
        plotFN = fn.parent / (fn.stem + "_picture.png")
        print("writing", plotFN)
        fg.savefig(plotFN, bbox_inches="tight", dpi=150)
